# coding=utf-8
# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import functools
import json
import os
import subprocess
from collections import OrderedDict, defaultdict, namedtuple
from textwrap import dedent

from pex import dist_metadata
from pex.common import AtomicDirectory, atomic_directory, safe_mkdtemp
from pex.distribution_target import DistributionTarget
from pex.interpreter import spawn_python_job
from pex.jobs import SpawnedJob, execute_parallel
from pex.orderedset import OrderedSet
from pex.pex_info import PexInfo
from pex.pip import get_pip
from pex.platforms import Platform
from pex.requirements import local_project_from_requirement, local_projects_from_requirement_file
from pex.third_party.packaging.markers import Marker
from pex.third_party.pkg_resources import Distribution, Environment, Requirement
from pex.tracer import TRACER
from pex.util import CacheHelper


class Untranslateable(Exception):
  pass


class Unsatisfiable(Exception):
  pass


class ResolvedDistribution(namedtuple('ResolvedDistribution',
                                      ['target', 'requirement', 'distribution'])):
  """A distribution target, requirement and the resolved distribution that satisfies them both."""

  def __new__(cls, target, requirement, distribution):
    assert isinstance(target, DistributionTarget)
    assert isinstance(requirement, Requirement)
    assert isinstance(distribution, Distribution)
    return super(ResolvedDistribution, cls).__new__(cls, target, requirement, distribution)


class DistributionRequirements(object):
  class Request(namedtuple('DistributionRequirementsRequest', ['target', 'distributions'])):
    def spawn_calculation(self):
      search_path = [dist.location for dist in self.distributions]

      program = dedent("""
        import json
        import sys
        from collections import defaultdict
        from pkg_resources import Environment


        env = Environment(search_path={search_path!r})
        dependency_requirements = []
        for key in env:
          for dist in env[key]:
            dependency_requirements.extend(str(req) for req in dist.requires())
        json.dump(dependency_requirements, sys.stdout)
      """.format(search_path=search_path))

      job = spawn_python_job(
        args=['-c', program],
        stdout=subprocess.PIPE,
        interpreter=self.target.get_interpreter(),
        expose=['setuptools']
      )
      return SpawnedJob.stdout(job=job, result_func=self._markers_by_requirement)

    @staticmethod
    def _markers_by_requirement(stdout):
      dependency_requirements = json.loads(stdout.decode('utf-8'))
      markers_by_req_key = defaultdict(OrderedSet)
      for requirement in dependency_requirements:
        req = Requirement.parse(requirement)
        if req.marker:
          markers_by_req_key[req.key].add(req.marker)
      return markers_by_req_key

  @classmethod
  def merged(cls, markers_by_requirement_key_iter):
    markers_by_requirement_key = defaultdict(OrderedSet)
    for distribution_markers in markers_by_requirement_key_iter:
      for requirement, markers in distribution_markers.items():
        markers_by_requirement_key[requirement].update(markers)
    return cls(markers_by_requirement_key)

  def __init__(self, markers_by_requirement_key):
    self._markers_by_requirement_key = markers_by_requirement_key

  def to_requirement(self, dist):
    req = dist.as_requirement()

    markers = OrderedSet()

    # Here we map any wheel python requirement to the equivalent environment marker:
    # See:
    # + https://www.python.org/dev/peps/pep-0345/#requires-python
    # + https://www.python.org/dev/peps/pep-0508/#environment-markers
    python_requires = dist_metadata.requires_python(dist)
    if python_requires:
      markers.update(
        Marker(python_version)
        for python_version in sorted(
          'python_version {operator} {version!r}'.format(
            operator=specifier.operator,
            version=specifier.version
          ) for specifier in python_requires
        )
      )

    markers.update(self._markers_by_requirement_key.get(req.key, ()))

    if not markers:
      return req

    if len(markers) == 1:
      marker = next(iter(markers))
      req.marker = marker
      return req

    # We may have resolved with multiple paths to the dependency represented by dist and at least
    # two of those paths had (different) conditional requirements for dist based on environment
    # marker predicates. In that case, since the pip resolve succeeded, the implication is that the
    # environment markers are compatible; i.e.: their intersection selects the target interpreter.
    # Here we make that intersection explicit.
    # See: https://www.python.org/dev/peps/pep-0508/#grammar
    marker = ' and '.join('({})'.format(marker) for marker in markers)
    return Requirement.parse('{}; {}'.format(req, marker))


def parsed_platform(platform=None):
  """Parse the given platform into a `Platform` object.

  Unlike `Platform.create`, this function supports the special platform of 'current' or `None`. This
  maps to the platform of any local python interpreter.

  :param platform: The platform string to parse. If `None` or 'current', return `None`. If already a
                   `Platform` object, return it.
  :type platform: str or :class:`Platform`
  :return: The parsed platform or `None` for the current platform.
  :rtype: :class:`Platform` or :class:`NoneType`
  """
  return Platform.create(platform) if platform and platform != 'current' else None


class ResolveResult(namedtuple('ResolveResult', ['target', 'download_dir'])):
  @staticmethod
  def _is_wheel(path):
    return os.path.isfile(path) and path.endswith('.whl')

  def _iter_distribution_paths(self):
    if not os.path.exists(self.download_dir):
      return
    for distribution in os.listdir(self.download_dir):
      yield os.path.join(self.download_dir, distribution)

  def build_requests(self):
    for distribution_path in self._iter_distribution_paths():
      if not self._is_wheel(distribution_path):
        yield BuildRequest.create(target=self.target, source_path=distribution_path)

  def install_requests(self):
    for distribution_path in self._iter_distribution_paths():
      if self._is_wheel(distribution_path):
        yield InstallRequest.create(target=self.target, wheel_path=distribution_path)


class BuildRequest(namedtuple('BuildRequest', ['target', 'source_path', 'fingerprint'])):
  @classmethod
  def create(cls, target, source_path):
    hasher = CacheHelper.dir_hash if os.path.isdir(source_path) else CacheHelper.hash
    fingerprint = hasher(source_path)
    return cls(target=target, source_path=source_path, fingerprint=fingerprint)

  def result(self, dist_root):
    return BuildResult.from_request(self, dist_root=dist_root)


class BuildResult(namedtuple('BuildResult', ['request', 'atomic_dir'])):
  @classmethod
  def from_request(cls, build_request, dist_root):
    dist_dir = os.path.join(
      dist_root,
      'sdists' if os.path.isfile(build_request.source_path) else 'local_projects',
      os.path.basename(build_request.source_path),
      build_request.fingerprint,
      build_request.target.id
    )
    return cls(request=build_request, atomic_dir=AtomicDirectory(dist_dir))

  @property
  def is_built(self):
    return self.atomic_dir.is_finalized

  @property
  def build_dir(self):
    return self.atomic_dir.work_dir

  @property
  def dist_dir(self):
    return self.atomic_dir.target_dir

  def finalize_build(self):
    self.atomic_dir.finalize()
    for wheel in os.listdir(self.dist_dir):
      yield InstallRequest.create(self.request.target, os.path.join(self.dist_dir, wheel))


class InstallRequest(namedtuple('InstallRequest', ['target', 'wheel_path', 'fingerprint'])):
  @classmethod
  def create(cls, target, wheel_path):
    fingerprint = CacheHelper.hash(wheel_path)
    return cls(target=target, wheel_path=wheel_path, fingerprint=fingerprint)

  @property
  def wheel_file(self):
    return os.path.basename(self.wheel_path)

  def result(self, installation_root):
    return InstallResult.from_request(self, installation_root=installation_root)


class InstallResult(namedtuple('InstallResult', ['request', 'installation_root', 'atomic_dir'])):
  @classmethod
  def from_request(cls, install_request, installation_root):
    install_chroot = os.path.join(
      installation_root,
      install_request.fingerprint,
      install_request.wheel_file
    )
    return cls(
      request=install_request,
      installation_root=installation_root,
      atomic_dir=AtomicDirectory(install_chroot)
    )

  @property
  def is_installed(self):
    return self.atomic_dir.is_finalized

  @property
  def build_chroot(self):
    return self.atomic_dir.work_dir

  @property
  def install_chroot(self):
    return self.atomic_dir.target_dir

  def finalize_install(self, install_requests):
    self.atomic_dir.finalize()

    # The install_chroot is keyed by the hash of the wheel file (zip) we installed. Here we add a
    # key by the hash of the exploded wheel dir (the install_chroot). This latter key is used by
    # zipped PEXes at runtime to explode their wheel chroots to the filesystem. By adding the key
    # here we short-circuit the explode process for PEXes created and run on the same machine.
    #
    # From a clean cache after building a simple pex this looks like:
    # $ rm -rf ~/.pex
    # $ python -mpex -c pex -o /tmp/pex.pex .
    # $ tree -L 4 ~/.pex/
    # /home/jsirois/.pex/
    # ├── built_wheels
    # │   └── 1003685de2c3604dc6daab9540a66201c1d1f718
    # │       └── cp-38-cp38
    # │           └── pex-2.0.2-py2.py3-none-any.whl
    # └── installed_wheels
    #     ├── 2a594cef34d2e9109bad847358d57ac4615f81f4
    #     │   └── pex-2.0.2-py2.py3-none-any.whl
    #     │       ├── bin
    #     │       ├── pex
    #     │       └── pex-2.0.2.dist-info
    #     └── ae13cba3a8e50262f4d730699a11a5b79536e3e1
    #         └── pex-2.0.2-py2.py3-none-any.whl -> /home/jsirois/.pex/installed_wheels/2a594cef34d2e9109bad847358d57ac4615f81f4/pex-2.0.2-py2.py3-none-any.whl  # noqa
    #
    # 11 directories, 1 file
    #
    # And we see in the created pex, the runtime key that the layout above satisfies:
    # $ unzip -qc /tmp/pex.pex PEX-INFO | jq .distributions
    # {
    #   "pex-2.0.2-py2.py3-none-any.whl": "ae13cba3a8e50262f4d730699a11a5b79536e3e1"
    # }
    #
    # When the pex is run, the runtime key is followed to the build time key, avoiding re-unpacking
    # the wheel:
    # $ PEX_VERBOSE=1 /tmp/pex.pex --version
    # pex: Found site-library: /usr/lib/python3.8/site-packages
    # pex: Tainted path element: /usr/lib/python3.8/site-packages
    # pex: Scrubbing from user site: /home/jsirois/.local/lib/python3.8/site-packages
    # pex: Scrubbing from site-packages: /usr/lib/python3.8/site-packages
    # pex: Activating PEX virtual environment from /tmp/pex.pex: 9.1ms
    # pex: Bootstrap complete, performing final sys.path modifications...
    # pex: PYTHONPATH contains:
    # pex:     /tmp/pex.pex
    # pex:   * /usr/lib/python38.zip
    # pex:     /usr/lib/python3.8
    # pex:     /usr/lib/python3.8/lib-dynload
    # pex:     /home/jsirois/.pex/installed_wheels/2a594cef34d2e9109bad847358d57ac4615f81f4/pex-2.0.2-py2.py3-none-any.whl  # noqa
    # pex:   * /tmp/pex.pex/.bootstrap
    # pex:   * - paths that do not exist or will be imported via zipimport
    # pex.pex 2.0.2
    #
    wheel_dir_hash = CacheHelper.dir_hash(self.install_chroot)
    runtime_key_dir = os.path.join(self.installation_root, wheel_dir_hash)
    with atomic_directory(runtime_key_dir) as work_dir:
      if work_dir:
        os.symlink(self.install_chroot, os.path.join(work_dir, self.request.wheel_file))

    return self._iter_requirements_requests(install_requests)

  def _iter_requirements_requests(self, install_requests):
    if self.is_installed:
      # N.B.: Direct snip from the Environment docs:
      #
      #  You may explicitly set `platform` (and/or `python`) to ``None`` if you
      #  wish to map *all* distributions, not just those compatible with the
      #  running platform or Python version.
      #
      # Since our requested target may be foreign, we make sure find all distributions installed by
      # explicitly setting both `python` and `platform` to `None`.
      environment = Environment(search_path=[self.install_chroot], python=None, platform=None)

      distributions = []
      for dist_project_name in environment:
        distributions.extend(environment[dist_project_name])

      for install_request in install_requests:
        yield DistributionRequirements.Request(
          target=install_request.target,
          distributions=distributions
        )


class ResolveRequest(object):
  def __init__(self,
               targets,
               requirements=None,
               requirement_files=None,
               constraint_files=None,
               allow_prereleases=False,
               transitive=True,
               indexes=None,
               find_links=None,
               cache=None,
               build=True,
               use_wheel=True,
               compile=False,
               manylinux=None,
               max_parallel_jobs=None):

    self._targets = targets
    self._requirements = requirements
    self._requirement_files = requirement_files
    self._constraint_files = constraint_files
    self._allow_prereleases = allow_prereleases
    self._transitive = transitive
    self._indexes = indexes
    self._find_links = find_links
    self._cache = cache
    self._build = build
    self._use_wheel = use_wheel
    self._compile = compile
    self._manylinux = manylinux
    self._max_parallel_jobs = max_parallel_jobs

  def _iter_local_projects(self):
    if self._requirements:
      for req in self._requirements:
        local_project = local_project_from_requirement(req)
        if local_project:
          for target in self._targets:
            yield BuildRequest.create(target=target, source_path=local_project)

    if self._requirement_files:
      for requirement_file in self._requirement_files:
        for local_project in local_projects_from_requirement_file(requirement_file):
          for target in self._targets:
            yield BuildRequest.create(target=target, source_path=local_project)

  def _run_parallel(self, inputs, spawn_func, raise_type):
    for result in execute_parallel(
        inputs=inputs,
        spawn_func=spawn_func,
        raise_type=raise_type,
        max_jobs=self._max_parallel_jobs
    ):
      yield result

  def _spawn_resolve(self, resolved_dists_dir, target):
    download_dir = os.path.join(resolved_dists_dir, target.id)
    download_job = get_pip().spawn_download_distributions(
      download_dir=download_dir,
      requirements=self._requirements,
      requirement_files=self._requirement_files,
      constraint_files=self._constraint_files,
      allow_prereleases=self._allow_prereleases,
      transitive=self._transitive,
      target=target,
      indexes=self._indexes,
      find_links=self._find_links,
      cache=self._cache,
      build=self._build,
      manylinux=self._manylinux,
      use_wheel=self._use_wheel
    )
    return SpawnedJob.wait(job=download_job, result=ResolveResult(target, download_dir))

  def _categorize_build_requests(self, build_requests, dist_root):
    unsatisfied_build_requests = []
    install_requests = []
    for build_request in build_requests:
      build_result = build_request.result(dist_root)
      if not build_result.is_built:
        TRACER.log('Building {} to {}'.format(build_request.source_path, build_result.dist_dir))
        unsatisfied_build_requests.append(build_request)
      else:
        TRACER.log('Using cached build of {} at {}'
                   .format(build_request.source_path, build_result.dist_dir))
        install_requests.extend(build_result.finalize_build())
    return unsatisfied_build_requests, install_requests

  def _spawn_wheel_build(self, built_wheels_dir, build_request):
    build_result = build_request.result(built_wheels_dir)
    build_job = get_pip().spawn_build_wheels(
      distributions=[build_request.source_path],
      wheel_dir=build_result.build_dir,
      cache=self._cache,
      indexes=self._indexes,
      find_links=self._find_links,
      interpreter=build_request.target.get_interpreter()
    )
    return SpawnedJob.wait(job=build_job, result=build_result)

  def _categorize_install_requests(self, install_requests, installed_wheels_dir):
    unsatisfied_install_requests = []
    install_results = []
    for install_request in install_requests:
      install_result = install_request.result(installed_wheels_dir)
      if not install_result.is_installed:
        TRACER.log('Installing {} in {}'
                   .format(install_request.wheel_path, install_result.install_chroot))
        unsatisfied_install_requests.append(install_request)
      else:
        TRACER.log('Using cached installation of {} at {}'
                   .format(install_request.wheel_file, install_result.install_chroot))
        install_results.append(install_result)
    return unsatisfied_install_requests, install_results

  def _spawn_install(self, installed_wheels_dir, install_request):
    install_result = install_request.result(installed_wheels_dir)
    install_job = get_pip().spawn_install_wheel(
      wheel=install_request.wheel_path,
      install_dir=install_result.build_chroot,
      compile=self._compile,
      cache=self._cache,
      target=install_request.target
    )
    return SpawnedJob.wait(job=install_job, result=install_result)

  def resolve_distributions(self, ignore_errors=False):
    # This method has four stages:
    # 1. Resolve sdists and wheels.
    # 2. Build local projects and sdists.
    # 3. Install wheels in individual chroots.
    # 4. Calculate the final resolved requirements.
    #
    # You'd think we might be able to just pip install all the requirements, but pexes can be
    # multi-platform / multi-interpreter, in which case only a subset of distributions resolved into
    # the PEX should be activated for the runtime interpreter. Sometimes there are platform specific
    # wheels and sometimes python version specific dists (backports being the common case). As such,
    # we need to be able to add each resolved distribution to the `sys.path` individually
    # (`PEXEnvironment` handles this selective activation at runtime). Since pip install only
    # accepts a single location to install all resolved dists, that won't work.
    #
    # This means we need to separately resolve all distributions, then install each in their own
    # chroot. To do this we use `pip download` for the resolve and download of all needed
    # distributions and then `pip install` to install each distribution in its own chroot.
    #
    # As a complicating factor, the runtime activation scheme relies on PEP 425 tags; i.e.: wheel
    # names. Some requirements are only available or applicable in source form - either via sdist,
    # VCS URL or local projects. As such we need to insert a `pip wheel` step to generate wheels for
    # all requirements resolved in source form via `pip download` / inspection of requirements to
    # discover those that are local directories (local setup.py or pyproject.toml python projects).
    #
    # Finally, we must calculate the pinned requirement corresponding to each distribution we
    # resolved along with any environment markers that control which runtime environments the
    # requirement should be activated in.

    if not self._requirements and not self._requirement_files:
      # Nothing to resolve.
      return []

    workspace = safe_mkdtemp()
    cache = self._cache or workspace

    resolved_dists_dir = os.path.join(workspace, 'resolved_dists')
    spawn_resolve = functools.partial(self._spawn_resolve, resolved_dists_dir)
    to_resolve = self._targets

    built_wheels_dir = os.path.join(cache, 'built_wheels')
    spawn_wheel_build = functools.partial(self._spawn_wheel_build, built_wheels_dir)
    to_build = list(self._iter_local_projects())

    installed_wheels_dir = os.path.join(cache, PexInfo.INSTALL_CACHE)
    spawn_install = functools.partial(self._spawn_install, installed_wheels_dir)
    to_install = []

    to_calculate_requirements_for = []

    # 1. Resolve sdists and wheels.
    with TRACER.timed('Resolving for:\n  {}'.format('\n  '.join(map(str, to_resolve)))):
      for resolve_result in self._run_parallel(inputs=to_resolve,
                                               spawn_func=spawn_resolve,
                                               raise_type=Unsatisfiable):
        to_build.extend(resolve_result.build_requests())
        to_install.extend(resolve_result.install_requests())

    if not any((to_build, to_install)):
      # Nothing to build or install.
      return []

    # 2. Build local projects and sdists.
    if to_build:
      with TRACER.timed('Building distributions for:\n  {}'
                        .format('\n  '.join(map(str, to_build)))):

        build_requests, install_requests = self._categorize_build_requests(
          build_requests=to_build,
          dist_root=built_wheels_dir
        )
        to_install.extend(install_requests)

        for build_result in self._run_parallel(inputs=build_requests,
                                               spawn_func=spawn_wheel_build,
                                               raise_type=Untranslateable):
          to_install.extend(build_result.finalize_build())

    # 3. Install wheels in individual chroots.

    # Dedup by wheel name; e.g.: only install universal wheels once even though they'll get
    # downloaded / built for each interpreter or platform.
    install_requests_by_wheel_file = OrderedDict()
    for install_request in to_install:
      install_requests = install_requests_by_wheel_file.setdefault(install_request.wheel_file, [])
      install_requests.append(install_request)

    representative_install_requests = [
      requests[0] for requests in install_requests_by_wheel_file.values()
    ]

    def add_requirements_requests(install_result):
      install_requests = install_requests_by_wheel_file[install_result.request.wheel_file]
      to_calculate_requirements_for.extend(install_result.finalize_install(install_requests))

    with TRACER.timed('Installing:\n  {}'
                      .format('\n  '.join(map(str, representative_install_requests)))):

      install_requests, install_results = self._categorize_install_requests(
        install_requests=representative_install_requests,
        installed_wheels_dir=installed_wheels_dir
      )
      for install_result in install_results:
        add_requirements_requests(install_result)

      for install_result in self._run_parallel(inputs=install_requests,
                                               spawn_func=spawn_install,
                                               raise_type=Untranslateable):
        add_requirements_requests(install_result)

    # 4. Calculate the final resolved requirements.
    with TRACER.timed('Calculating resolved requirements for:\n  {}'
                      .format('\n  '.join(map(str, to_calculate_requirements_for)))):
      distribution_requirements = DistributionRequirements.merged(
        self._run_parallel(
          inputs=to_calculate_requirements_for,
          spawn_func=DistributionRequirements.Request.spawn_calculation,
          raise_type=Untranslateable
        )
      )

    resolved_distributions = OrderedSet()
    for requirements_request in to_calculate_requirements_for:
      for distribution in requirements_request.distributions:
        resolved_distributions.add(
          ResolvedDistribution(
            target=requirements_request.target,
            requirement=distribution_requirements.to_requirement(distribution),
            distribution=distribution
          )
        )

    if not ignore_errors and self._transitive:
      self._check_resolve(resolved_distributions)
    return resolved_distributions

  def _check_resolve(self, resolved_distributions):
    resolved_distribution_by_key = OrderedDict(
      (resolved_distribution.requirement.key, resolved_distribution)
      for resolved_distribution in resolved_distributions
    )

    unsatisfied = []
    for resolved_distribution in resolved_distribution_by_key.values():
      dist = resolved_distribution.distribution
      target = resolved_distribution.target
      for requirement in dist.requires():
        if not target.requirement_applies(requirement):
          continue

        resolved_requirement_dist = resolved_distribution_by_key.get(requirement.key)
        if not resolved_requirement_dist:
          unsatisfied.append(
            '{dist} requires {requirement} but no version was resolved'.format(
              dist=dist.as_requirement(),
              requirement=requirement
            )
          )
        else:
          resolved_dist = resolved_requirement_dist.distribution
          if resolved_dist not in requirement:
            unsatisfied.append(
              '{dist} requires {requirement} but {resolved_dist} was resolved'.format(
                dist=dist.as_requirement(),
                requirement=requirement,
                resolved_dist=resolved_dist
              )
            )

    if unsatisfied:
      raise Unsatisfiable(
        'Failed to resolve compatible distributions:\n{failures}'.format(
          failures='\n'.join('{index}: {failure}'.format(index=index + 1, failure=failure)
                             for index, failure in enumerate(unsatisfied))
        )
      )


def resolve(requirements=None,
            requirement_files=None,
            constraint_files=None,
            allow_prereleases=False,
            transitive=True,
            interpreter=None,
            platform=None,
            indexes=None,
            find_links=None,
            cache=None,
            build=True,
            use_wheel=True,
            compile=False,
            manylinux=None,
            max_parallel_jobs=None,
            ignore_errors=False):
  """Produce all distributions needed to meet all specified requirements.

  :keyword requirements: A sequence of requirement strings.
  :type requirements: list of str
  :keyword requirement_files: A sequence of requirement file paths.
  :type requirement_files: list of str
  :keyword constraint_files: A sequence of constraint file paths.
  :type constraint_files: list of str
  :keyword bool allow_prereleases: Whether to include pre-release and development versions when
    resolving requirements. Defaults to ``False``, but any requirements that explicitly request
    prerelease or development versions will override this setting.
  :keyword bool transitive: Whether to resolve transitive dependencies of requirements.
    Defaults to ``True``.
  :keyword interpreter: The interpreter to use for building distributions and for testing
    distribution compatibility. Defaults to the current interpreter.
  :type interpreter: :class:`pex.interpreter.PythonInterpreter`
  :keyword str platform: The exact target platform to resolve distributions for. If ``None`` or
    ``'current'``, resolve for distributions appropriate for `interpreter`.
  :keyword indexes: A list of urls or paths pointing to PEP 503 compliant repositories to search for
    distributions. Defaults to ``None`` which indicates to use the default pypi index. To turn off
    use of all indexes, pass an empty list.
  :type indexes: list of str
  :keyword find_links: A list or URLs, paths to local html files or directory paths. If URLs or
    local html file paths, these are parsed for links to distributions. If a local directory path,
    its listing is used to discover distributons.
  :type find_links: list of str
  :keyword str cache: A directory path to use to cache distributions locally.
  :keyword bool build: Whether to allow building source distributions when no wheel is found.
    Defaults to ``True``.
  :keyword bool use_wheel: Whether to allow resolution of pre-built wheel distributions.
    Defaults to ``True``.
  :keyword bool compile: Whether to pre-compile resolved distribution python sources.
    Defaults to ``False``.
  :keyword str manylinux: The upper bound manylinux standard to support when targeting foreign linux
    platforms. Defaults to ``None``.
  :keyword int max_parallel_jobs: The maximum number of parallel jobs to use when resolving,
    building and installing distributions in a resolve. Defaults to the number of CPUs available.
  :keyword bool ignore_errors: Whether to ignore resolution solver errors. Defaults to ``False``.
  :returns: List of :class:`ResolvedDistribution` instances meeting ``requirements``.
  :raises Unsatisfiable: If ``requirements`` is not transitively satisfiable.
  :raises Untranslateable: If no compatible distributions could be acquired for
    a particular requirement.
  """

  target = DistributionTarget(interpreter=interpreter, platform=parsed_platform(platform))

  resolve_request = ResolveRequest(targets=[target],
                                   requirements=requirements,
                                   requirement_files=requirement_files,
                                   constraint_files=constraint_files,
                                   allow_prereleases=allow_prereleases,
                                   transitive=transitive,
                                   indexes=indexes,
                                   find_links=find_links,
                                   cache=cache,
                                   build=build,
                                   use_wheel=use_wheel,
                                   compile=compile,
                                   manylinux=manylinux,
                                   max_parallel_jobs=max_parallel_jobs)

  return list(resolve_request.resolve_distributions(ignore_errors=ignore_errors))


def resolve_multi(requirements=None,
                  requirement_files=None,
                  constraint_files=None,
                  allow_prereleases=False,
                  transitive=True,
                  interpreters=None,
                  platforms=None,
                  indexes=None,
                  find_links=None,
                  cache=None,
                  build=True,
                  use_wheel=True,
                  compile=False,
                  manylinux=None,
                  max_parallel_jobs=None,
                  ignore_errors=False):
  """A generator function that produces all distributions needed to meet `requirements`
  for multiple interpreters and/or platforms.

  :keyword requirements: A sequence of requirement strings.
  :type requirements: list of str
  :keyword requirement_files: A sequence of requirement file paths.
  :type requirement_files: list of str
  :keyword constraint_files: A sequence of constraint file paths.
  :type constraint_files: list of str
  :keyword bool allow_prereleases: Whether to include pre-release and development versions when
    resolving requirements. Defaults to ``False``, but any requirements that explicitly request
    prerelease or development versions will override this setting.
  :keyword bool transitive: Whether to resolve transitive dependencies of requirements.
    Defaults to ``True``.
  :keyword interpreters: The interpreters to use for building distributions and for testing
    distribution compatibility. Defaults to the current interpreter.
  :type interpreters: list of :class:`pex.interpreter.PythonInterpreter`
  :keyword platforms: An iterable of PEP425-compatible platform strings to resolve distributions
    for. If ``None`` (the default) or an empty iterable, use the platforms of the given
    interpreters.
  :type platforms: list of str
  :keyword indexes: A list of urls or paths pointing to PEP 503 compliant repositories to search for
    distributions. Defaults to ``None`` which indicates to use the default pypi index. To turn off
    use of all indexes, pass an empty list.
  :type indexes: list of str
  :keyword find_links: A list or URLs, paths to local html files or directory paths. If URLs or
    local html file paths, these are parsed for links to distributions. If a local directory path,
    its listing is used to discover distributons.
  :type find_links: list of str
  :keyword str cache: A directory path to use to cache distributions locally.
  :keyword bool build: Whether to allow building source distributions when no wheel is found.
    Defaults to ``True``.
  :keyword bool use_wheel: Whether to allow resolution of pre-built wheel distributions.
    Defaults to ``True``.
  :keyword bool compile: Whether to pre-compile resolved distribution python sources.
    Defaults to ``False``.
  :keyword str manylinux: The upper bound manylinux standard to support when targeting foreign linux
    platforms. Defaults to ``None``.
  :keyword int max_parallel_jobs: The maximum number of parallel jobs to use when resolving,
    building and installing distributions in a resolve. Defaults to the number of CPUs available.
  :keyword bool ignore_errors: Whether to ignore resolution solver errors. Defaults to ``False``.
  :returns: List of :class:`ResolvedDistribution` instances meeting ``requirements``.
  :raises Unsatisfiable: If ``requirements`` is not transitively satisfiable.
  :raises Untranslateable: If no compatible distributions could be acquired for
    a particular requirement.
  """

  parsed_platforms = [parsed_platform(platform) for platform in platforms] if platforms else []

  def iter_targets():
    if not interpreters and not parsed_platforms:
      # No specified targets, so just build for the current interpreter (on the current platform).
      yield DistributionTarget.current()
      return

    if interpreters:
      for interpreter in interpreters:
        # Build for the specified local interpreters (on the current platform).
        yield DistributionTarget.for_interpreter(interpreter)

    if parsed_platforms:
      for platform in parsed_platforms:
        if platform is not None or not interpreters:
          # 1. Build for specific platforms.
          # 2. Build for the current platform (None) only if not done already (ie: no intepreters
          #    were specified).
          yield DistributionTarget.for_platform(platform)

  resolve_request = ResolveRequest(targets=list(iter_targets()),
                                   requirements=requirements,
                                   requirement_files=requirement_files,
                                   constraint_files=constraint_files,
                                   allow_prereleases=allow_prereleases,
                                   transitive=transitive,
                                   indexes=indexes,
                                   find_links=find_links,
                                   cache=cache,
                                   build=build,
                                   use_wheel=use_wheel,
                                   compile=compile,
                                   manylinux=manylinux,
                                   max_parallel_jobs=max_parallel_jobs)

  return list(resolve_request.resolve_distributions(ignore_errors=ignore_errors))



# coding=utf-8
# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os
from collections import deque
from textwrap import dedent

from pex import third_party
from pex.common import atomic_directory
from pex.compatibility import urlparse
from pex.distribution_target import DistributionTarget
from pex.jobs import Job
from pex.third_party import isolated
from pex.tracer import TRACER
from pex.variables import ENV


class Pip(object):
  @classmethod
  def create(cls, path):
    """Creates a pip tool with PEX isolation at path.

    :param str path: The path to build the pip tool pex at.
    """
    pip_pex_path = os.path.join(path, isolated().pex_hash)
    with atomic_directory(pip_pex_path) as chroot:
      if chroot is not None:
        from pex.pex_builder import PEXBuilder

        isolated_pip_builder = PEXBuilder(path=chroot)
        pythonpath = third_party.expose(['pip', 'setuptools', 'wheel'])
        isolated_pip_environment = third_party.pkg_resources.Environment(search_path=pythonpath)
        for dist_name in isolated_pip_environment:
          for dist in isolated_pip_environment[dist_name]:
            isolated_pip_builder.add_dist_location(dist=dist.location)
        with open(os.path.join(isolated_pip_builder.path(), 'run_pip.py'), 'w') as fp:
          fp.write(dedent("""\
            import os
            import runpy
            import sys


            # Propagate un-vendored setuptools to pip for any legacy setup.py builds it needs to
            # perform.
            os.environ['__PEX_UNVENDORED__'] = '1'
            os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)

            runpy.run_module('pip', run_name='__main__')
          """))
        isolated_pip_builder.set_executable(fp.name)
        isolated_pip_builder.freeze()

    return cls(pip_pex_path)

  def __init__(self, pip_pex_path):
    self._pip_pex_path = pip_pex_path

  def _spawn_pip_isolated(self, args, cache=None, interpreter=None):
    pip_args = [
      # We vendor the version of pip we want so pip should never check for updates.
      '--disable-pip-version-check',

      # Don't read pip configuration files like `~/.config/pip/pip.conf`.
      '--isolated',

      # If we want to warn about a version of python we support, we should do it, not pip.
      '--no-python-version-warning'
    ]

    # The max pip verbosity is -vvv and for pex it's -vvvvvvvvv; so we scale down by a factor of 3.
    pex_verbosity = ENV.PEX_VERBOSE
    pip_verbosity = pex_verbosity // 3
    if pip_verbosity > 0:
      pip_args.append('-{}'.format('v' * pip_verbosity))
    else:
      pip_args.append('-q')

    if cache:
      pip_args.extend(['--cache-dir', cache])
    else:
      pip_args.append('--no-cache-dir')

    command = pip_args + args
    with ENV.strip().patch(PEX_ROOT=cache or ENV.PEX_ROOT, PEX_VERBOSE=str(pex_verbosity)) as env:
      # Guard against API calls from environment with ambient PYTHONPATH preventing pip PEX
      # bootstrapping. See: https://github.com/pantsbuild/pex/issues/892
      pythonpath = env.pop('PYTHONPATH', None)
      if pythonpath:
        TRACER.log('Scrubbed PYTHONPATH={} from the pip PEX environment.'.format(pythonpath), V=3)

      from pex.pex import PEX
      pip = PEX(pex=self._pip_pex_path, interpreter=interpreter)
      return Job(
        command=pip.cmdline(command),
        process=pip.run(
          args=command,
          env=env,
          blocking=False
        )
      )

  def _calculate_package_index_options(self, indexes=None, find_links=None):
    trusted_hosts = []

    def maybe_trust_insecure_host(url):
      url_info = urlparse.urlparse(url)
      if 'http' == url_info.scheme:
        # Implicitly trust explicitly asked for http indexes and find_links repos instead of
        # requiring seperate trust configuration.
        trusted_hosts.append(url_info.netloc)
      return url

    # N.B.: We interpret None to mean accept pip index defaults, [] to mean turn off all index use.
    if indexes is not None:
      if len(indexes) == 0:
        yield '--no-index'
      else:
        all_indexes = deque(indexes)
        yield '--index-url'
        yield maybe_trust_insecure_host(all_indexes.popleft())
        if all_indexes:
          for extra_index in all_indexes:
            yield '--extra-index-url'
            yield maybe_trust_insecure_host(extra_index)

    if find_links:
      for find_link_url in find_links:
        yield '--find-links'
        yield maybe_trust_insecure_host(find_link_url)

    for trusted_host in trusted_hosts:
      yield '--trusted-host'
      yield trusted_host

  def spawn_download_distributions(self,
                                   download_dir,
                                   requirements=None,
                                   requirement_files=None,
                                   constraint_files=None,
                                   allow_prereleases=False,
                                   transitive=True,
                                   target=None,
                                   indexes=None,
                                   find_links=None,
                                   cache=None,
                                   build=True,
                                   manylinux=None,
                                   use_wheel=True):

    target = target or DistributionTarget.current()

    platform = target.get_platform()
    if not use_wheel:
      if not build:
        raise ValueError('Cannot both ignore wheels (use_wheel=False) and refrain from building '
                         'distributions (build=False).')
      elif target.is_foreign:
        raise ValueError('Cannot ignore wheels (use_wheel=False) when resolving for a foreign '
                         'platform: {}'.format(platform))

    download_cmd = ['download', '--dest', download_dir]
    package_index_options = self._calculate_package_index_options(
      indexes=indexes,
      find_links=find_links
    )
    download_cmd.extend(package_index_options)

    if target.is_foreign:
      # We're either resolving for a different host / platform or a different interpreter for the
      # current platform that we have no access to; so we need to let pip know and not otherwise
      # pickup platform info from the interpreter we execute pip with.
      if manylinux and platform.platform.startswith('linux'):
        download_cmd.extend(['--platform', platform.platform.replace('linux', manylinux, 1)])
      download_cmd.extend(['--platform', platform.platform])
      download_cmd.extend(['--implementation', platform.impl])
      download_cmd.extend(['--python-version', platform.version])
      download_cmd.extend(['--abi', platform.abi])

    if target.is_foreign or not build:
      download_cmd.extend(['--only-binary', ':all:'])

    if not use_wheel:
      download_cmd.extend(['--no-binary', ':all:'])

    if allow_prereleases:
      download_cmd.append('--pre')

    if not transitive:
      download_cmd.append('--no-deps')

    if requirement_files:
      for requirement_file in requirement_files:
        download_cmd.extend(['--requirement', requirement_file])

    if constraint_files:
      for constraint_file in constraint_files:
        download_cmd.extend(['--constraint', constraint_file])

    download_cmd.extend(requirements)

    return self._spawn_pip_isolated(download_cmd, cache=cache, interpreter=target.get_interpreter())

  def spawn_build_wheels(self,
                         distributions,
                         wheel_dir,
                         interpreter=None,
                         indexes=None,
                         find_links=None,
                         cache=None):

    wheel_cmd = ['wheel', '--no-deps', '--wheel-dir', wheel_dir]

    # If the build is PEP-517 compliant it may need to resolve build requirements.
    wheel_cmd.extend(self._calculate_package_index_options(indexes=indexes, find_links=find_links))

    wheel_cmd.extend(distributions)
    return self._spawn_pip_isolated(wheel_cmd, cache=cache, interpreter=interpreter)

  def spawn_install_wheel(self,
                          wheel,
                          install_dir,
                          compile=False,
                          cache=None,
                          target=None):

    target = target or DistributionTarget.current()

    install_cmd = [
      'install',
      '--no-deps',
      '--no-index',
      '--only-binary', ':all:',
      '--target', install_dir
    ]

    interpreter = target.get_interpreter()
    if target.is_foreign:
      if compile:
        raise ValueError('Cannot compile bytecode for {} using {} because the wheel has a foreign '
                         'platform.'.format(wheel, interpreter))

      # We're installing a wheel for a foreign platform. This is just an unpacking operation though;
      # so we don't actually need to perform it with a target platform compatible interpreter.
      install_cmd.append('--ignore-requires-python')

    install_cmd.append('--compile' if compile else '--no-compile')
    install_cmd.append(wheel)
    return self._spawn_pip_isolated(install_cmd, cache=cache, interpreter=interpreter)


_PIP = None


def get_pip():
  """Returns a lazily instantiated global Pip object that is safe for un-coordinated use."""
  global _PIP
  if _PIP is None:
    _PIP = Pip.create(path=os.path.join(ENV.PEX_ROOT, 'pip.pex'))
  return _PIP



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

import pytest

from pex import interpreter
from pex.compatibility import PY3
from pex.testing import PY27, PY35, ensure_python_interpreter

try:
  from mock import patch
except ImportError:
  from unittest.mock import patch


def tuple_from_version(version_string):
  return tuple(int(component) for component in version_string.split('.'))


class TestPythonInterpreter(object):
  def test_all_does_not_raise_with_empty_path_envvar(self):
    """ additionally, tests that the module does not raise at import """
    with patch.dict(os.environ, clear=True):
      if PY3:
        import importlib
        importlib.reload(interpreter)
      else:
        reload(interpreter)
      interpreter.PythonInterpreter.all()

  TEST_INTERPRETER1_VERSION = PY27
  TEST_INTERPRETER1_VERSION_TUPLE = tuple_from_version(TEST_INTERPRETER1_VERSION)

  TEST_INTERPRETER2_VERSION = PY35
  TEST_INTERPRETER2_VERSION_TUPLE = tuple_from_version(TEST_INTERPRETER2_VERSION)

  @pytest.fixture
  def test_interpreter1(self):
    return ensure_python_interpreter(self.TEST_INTERPRETER1_VERSION)

  @pytest.fixture
  def test_interpreter2(self):
    return ensure_python_interpreter(self.TEST_INTERPRETER2_VERSION)

  def test_interpreter_versioning(self, test_interpreter1):
    py_interpreter = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    assert py_interpreter.identity.version == self.TEST_INTERPRETER1_VERSION_TUPLE

  def test_interpreter_caching(self, test_interpreter1, test_interpreter2):
    py_interpreter1 = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    py_interpreter2 = interpreter.PythonInterpreter.from_binary(test_interpreter2)
    assert py_interpreter1 is not py_interpreter2
    assert py_interpreter2.identity.version == self.TEST_INTERPRETER2_VERSION_TUPLE

    py_interpreter3 = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    assert py_interpreter1 is py_interpreter3

  def test_nonexistent_interpreter(self):
    with pytest.raises(interpreter.PythonInterpreter.InterpreterNotFound):
      interpreter.PythonInterpreter.from_binary('/nonexistent/path')

  def test_binary_name_matching(self):
    valid_binary_names = (
      'jython',
      'pypy',
      'pypy-1.1',
      'python',
      'Python',
      'python2',
      'python2.7',
      'python2.7m',
      'python3',
      'python3.6',
      'python3.6m'
    )

    matches = interpreter.PythonInterpreter._matches_binary_name
    for name in valid_binary_names:
      assert matches(name), 'Expected {} to be valid binary name'.format(name)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

"""pex support for interacting with interpreters."""

from __future__ import absolute_import

import hashlib
import json
import os
import platform
import re
import subprocess
import sys
from textwrap import dedent

from pex import third_party
from pex.common import safe_rmtree
from pex.compatibility import string
from pex.executor import Executor
from pex.jobs import Job, SpawnedJob, execute_parallel
from pex.third_party.packaging import markers, tags
from pex.third_party.pkg_resources import Distribution, Requirement
from pex.tracer import TRACER
from pex.util import CacheHelper
from pex.variables import ENV


class PythonIdentity(object):
  class Error(Exception): pass
  class InvalidError(Error): pass
  class UnknownRequirement(Error): pass

  # TODO(wickman)  Support interpreter-specific versions, e.g. PyPy-2.2.1
  INTERPRETER_NAME_TO_HASHBANG = {
    'CPython': 'python%(major)d.%(minor)d',
    'Jython': 'jython',
    'PyPy': 'pypy',
    'IronPython': 'ipy',
  }

  ABBR_TO_INTERPRETER_NAME = {
    'pp': 'PyPy',
    'jy': 'Jython',
    'ip': 'IronPython',
    'cp': 'CPython',
  }

  @classmethod
  def get(cls):
    supported_tags = tuple(tags.sys_tags())
    preferred_tag = supported_tags[0]
    return cls(
      binary=sys.executable,
      python_tag=preferred_tag.interpreter,
      abi_tag=preferred_tag.abi,
      platform_tag=preferred_tag.platform,
      version=sys.version_info[:3],
      supported_tags=supported_tags,
      env_markers=markers.default_environment()
    )

  @classmethod
  def decode(cls, encoded):
    TRACER.log('creating PythonIdentity from encoded: %s' % encoded, V=9)
    values = json.loads(encoded)
    if len(values) != 7:
      raise cls.InvalidError("Invalid interpreter identity: %s" % encoded)

    supported_tags = values.pop('supported_tags')

    def iter_tags():
      for (interpreter, abi, platform) in supported_tags:
        yield tags.Tag(interpreter=interpreter, abi=abi, platform=platform)

    return cls(supported_tags=iter_tags(), **values)

  @classmethod
  def _find_interpreter_name(cls, python_tag):
    for abbr, interpreter in cls.ABBR_TO_INTERPRETER_NAME.items():
      if python_tag.startswith(abbr):
        return interpreter
    raise ValueError('Unknown interpreter: {}'.format(python_tag))

  def __init__(
      self,
      binary,
      python_tag,
      abi_tag,
      platform_tag,
      version,
      supported_tags,
      env_markers
  ):
    # N.B.: We keep this mapping to support historical values for `distribution` and `requirement`
    # properties.
    self._interpreter_name = self._find_interpreter_name(python_tag)

    self._binary = binary
    self._python_tag = python_tag
    self._abi_tag = abi_tag
    self._platform_tag = platform_tag
    self._version = tuple(version)
    self._supported_tags = tuple(supported_tags)
    self._env_markers = dict(env_markers)

  def encode(self):
    values = dict(
      binary=self._binary,
      python_tag=self._python_tag,
      abi_tag=self._abi_tag,
      platform_tag=self._platform_tag,
      version=self._version,
      supported_tags=[(tag.interpreter, tag.abi, tag.platform) for tag in self._supported_tags],
      env_markers=self._env_markers
    )
    return json.dumps(values)

  @property
  def binary(self):
    return self._binary

  @property
  def python_tag(self):
    return self._python_tag

  @property
  def abi_tag(self):
    return self._abi_tag

  @property
  def platform_tag(self):
    return self._platform_tag

  @property
  def version(self):
    return self._version

  @property
  def version_str(self):
    return '.'.join(map(str, self.version))

  @property
  def supported_tags(self):
    return self._supported_tags

  @property
  def env_markers(self):
    return dict(self._env_markers)

  @property
  def interpreter(self):
    return self._interpreter_name

  @property
  def requirement(self):
    return self.distribution.as_requirement()

  @property
  def distribution(self):
    return Distribution(project_name=self.interpreter, version=self.version_str)

  @classmethod
  def parse_requirement(cls, requirement, default_interpreter='CPython'):
    if isinstance(requirement, Requirement):
      return requirement
    elif isinstance(requirement, string):
      try:
        requirement = Requirement.parse(requirement)
      except ValueError:
        try:
          requirement = Requirement.parse('%s%s' % (default_interpreter, requirement))
        except ValueError:
          raise ValueError('Unknown requirement string: %s' % requirement)
      return requirement
    else:
      raise ValueError('Unknown requirement type: %r' % (requirement,))

  def matches(self, requirement):
    """Given a Requirement, check if this interpreter matches."""
    try:
      requirement = self.parse_requirement(requirement, self._interpreter_name)
    except ValueError as e:
      raise self.UnknownRequirement(str(e))
    return self.distribution in requirement

  def hashbang(self):
    hashbang_string = self.INTERPRETER_NAME_TO_HASHBANG.get(self._interpreter_name, 'CPython') % {
      'major': self._version[0],
      'minor': self._version[1],
      'patch': self._version[2],
    }
    return '#!/usr/bin/env %s' % hashbang_string

  @property
  def python(self):
    # return the python version in the format of the 'python' key for distributions
    # specifically, '2.7', '3.2', etc.
    return '%d.%d' % (self.version[0:2])

  def __str__(self):
    # N.B.: Kept as distinct from __repr__ to support legacy str(identity) used by Pants v1 when
    # forming cache locations.
    return '{interpreter_name}-{major}.{minor}.{patch}'.format(
      interpreter_name=self._interpreter_name,
      major=self._version[0],
      minor=self._version[1],
      patch=self._version[2]
    )

  def __repr__(self):
    return '{type}({binary!r}, {python_tag!r}, {abi_tag!r}, {platform_tag!r}, {version!r})'.format(
      type=self.__class__.__name__,
      binary=self._binary,
      python_tag=self._python_tag,
      abi_tag=self._abi_tag,
      platform_tag=self._platform_tag,
      version=self._version
    )

  def _tup(self):
    return self._binary, self._python_tag, self._abi_tag, self._platform_tag, self._version

  def __eq__(self, other):
    if type(other) is not type(self):
      return NotImplemented
    return self._tup() == other._tup()

  def __hash__(self):
    return hash(self._tup())


class PythonInterpreter(object):
  _REGEXEN = (
    re.compile(r'jython$'),

    # NB: OSX ships python binaries named Python so we allow for capital-P.
    re.compile(r'[Pp]ython$'),

    re.compile(r'python[23]$'),
    re.compile(r'python[23].[0-9]$'),

    # Some distributions include a suffix on the in the interpreter name, similar to PEP-3149
    # E.g. Gentoo has /usr/bin/python3.6m to indicate it was built with pymalloc
    re.compile(r'python[23].[0-9][a-z]$'),

    re.compile(r'pypy$'),
    re.compile(r'pypy-1.[0-9]$'),
  )

  _PYTHON_INTERPRETER_BY_NORMALIZED_PATH = {}

  @staticmethod
  def _normalize_path(path):
    return os.path.realpath(path)

  class Error(Exception): pass
  class IdentificationError(Error): pass
  class InterpreterNotFound(Error): pass

  @classmethod
  def get(cls):
    return cls.from_binary(sys.executable)

  @staticmethod
  def _paths(paths=None):
    return paths or os.getenv('PATH', '').split(os.pathsep)

  @classmethod
  def iter(cls, paths=None):
    """Iterate all interpreters found in `paths`.

    NB: The paths can either be directories to search for python binaries or the paths of python
    binaries themselves.

    :param paths: The paths to look for python interpreters; by default the `PATH`.
    :type paths: list str
    """
    return cls._filter(cls._find(cls._paths(paths=paths)))

  @classmethod
  def all(cls, paths=None):
    return list(cls.iter(paths=paths))

  @classmethod
  def _create_isolated_cmd(cls, binary, args=None, pythonpath=None, env=None):
    cmd = [binary]

    # Don't add the user site directory to `sys.path`.
    #
    # Additionally, it would be nice to pass `-S` to disable adding site-packages but unfortunately
    # some python distributions include portions of the standard library there.
    cmd.append('-s')

    env = cls._sanitized_environment(env=env)
    pythonpath = list(pythonpath or ())
    if pythonpath:
      env['PYTHONPATH'] = os.pathsep.join(pythonpath)
    else:
      # Turn off reading of PYTHON* environment variables.
      cmd.append('-E')

    if args:
      cmd.extend(args)

    rendered_command = ' '.join(cmd)
    if pythonpath:
      rendered_command = 'PYTHONPATH={} {}'.format(env['PYTHONPATH'], rendered_command)
    TRACER.log('Executing: {}'.format(rendered_command), V=3)

    return cmd, env

  @classmethod
  def _execute(cls, binary, args=None, pythonpath=None, env=None, stdin_payload=None, **kwargs):
    cmd, env = cls._create_isolated_cmd(binary, args=args, pythonpath=pythonpath, env=env)
    stdout, stderr = Executor.execute(cmd, stdin_payload=stdin_payload, env=env, **kwargs)
    return cmd, stdout, stderr

  INTERP_INFO_FILE = 'INTERP-INFO'

  @classmethod
  def _spawn_from_binary_external(cls, binary):
    def create_interpreter(stdout):
      identity = stdout.decode('utf-8').strip()
      if not identity:
        raise cls.IdentificationError('Could not establish identity of %s' % binary)
      return cls(PythonIdentity.decode(identity))

    # Part of the PythonInterpreter data are environment markers that depend on the current OS
    # release. That data can change when the OS is upgraded but (some of) the installed interpreters
    # remain the same. As such, include the OS in the hash structure for cached interpreters.
    os_digest = hashlib.sha1()
    for os_identifier in platform.release(), platform.version():
      os_digest.update(os_identifier.encode('utf-8'))
    os_hash = os_digest.hexdigest()

    interpreter_cache_dir = os.path.join(ENV.PEX_ROOT, 'interpreters')
    os_cache_dir = os.path.join(interpreter_cache_dir, os_hash)
    if os.path.isdir(interpreter_cache_dir) and not os.path.isdir(os_cache_dir):
      with TRACER.timed('GCing interpreter cache from prior OS version'):
        safe_rmtree(interpreter_cache_dir)

    interpreter_hash = CacheHelper.hash(binary)
    cache_dir = os.path.join(os_cache_dir, interpreter_hash)
    cache_file = os.path.join(cache_dir, cls.INTERP_INFO_FILE)
    if os.path.isfile(cache_file):
      try:
        with open(cache_file, 'rb') as fp:
          return SpawnedJob.completed(create_interpreter(fp.read()))
      except (IOError, OSError, cls.Error, PythonIdentity.Error):
        safe_rmtree(cache_dir)
        return cls._spawn_from_binary_external(binary)
    else:
      pythonpath = third_party.expose(['pex'])
      cmd, env = cls._create_isolated_cmd(
        binary,
        args=[
          '-c',
          dedent("""\
          import os
          import sys

          from pex.common import atomic_directory, safe_open
          from pex.interpreter import PythonIdentity


          encoded_identity = PythonIdentity.get().encode()
          sys.stdout.write(encoded_identity)
          with atomic_directory({cache_dir!r}) as cache_dir:
            if cache_dir:
              with safe_open(os.path.join(cache_dir, {info_file!r}), 'w') as fp:
                fp.write(encoded_identity)
          """.format(cache_dir=cache_dir, info_file=cls.INTERP_INFO_FILE))
        ],
        pythonpath=pythonpath
      )
      process = Executor.open_process(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      job = Job(command=cmd, process=process)
      return SpawnedJob.stdout(job, result_func=create_interpreter)

  @classmethod
  def _expand_path(cls, path):
    if os.path.isfile(path):
      return [path]
    elif os.path.isdir(path):
      return sorted(os.path.join(path, fn) for fn in os.listdir(path))
    return []

  @classmethod
  def from_env(cls, hashbang):
    """Resolve a PythonInterpreter as /usr/bin/env would.

    :param hashbang: A string, e.g. "python3.3" representing some binary on the $PATH.
    :return: the first matching interpreter found or `None`.
    :rtype: :class:`PythonInterpreter`
    """
    def hashbang_matches(fn):
      basefile = os.path.basename(fn)
      return hashbang == basefile

    for interpreter in cls._identify_interpreters(filter=hashbang_matches):
      return interpreter

  @classmethod
  def _spawn_from_binary(cls, binary):
    normalized_binary = cls._normalize_path(binary)
    if not os.path.exists(normalized_binary):
      raise cls.InterpreterNotFound(normalized_binary)

    # N.B.: The cache is written as the last step in PythonInterpreter instance initialization.
    cached_interpreter = cls._PYTHON_INTERPRETER_BY_NORMALIZED_PATH.get(normalized_binary)
    if cached_interpreter is not None:
      return SpawnedJob.completed(cached_interpreter)
    if normalized_binary == cls._normalize_path(sys.executable):
      current_interpreter = cls(PythonIdentity.get())
      return SpawnedJob.completed(current_interpreter)
    return cls._spawn_from_binary_external(normalized_binary)

  @classmethod
  def from_binary(cls, binary):
    """Create an interpreter from the given `binary`.

    :param str binary: The path to the python interpreter binary.
    :return: an interpreter created from the given `binary`.
    :rtype: :class:`PythonInterpreter`
    """
    return cls._spawn_from_binary(binary).await_result()

  @classmethod
  def _matches_binary_name(cls, path):
    basefile = os.path.basename(path)
    return any(matcher.match(basefile) is not None for matcher in cls._REGEXEN)

  @classmethod
  def _find(cls, paths):
    """Given a list of files or directories, try to detect python interpreters amongst them.

    Returns an iterator over PythonInterpreter objects.
    """
    return cls._identify_interpreters(filter=cls._matches_binary_name, paths=paths)

  @classmethod
  def _identify_interpreters(cls, filter, paths=None):
    def iter_candidates():
      for path in cls._paths(paths=paths):
        for fn in cls._expand_path(path):
          if filter(fn):
            yield fn

    return execute_parallel(inputs=list(iter_candidates()), spawn_func=cls._spawn_from_binary)

  @classmethod
  def _filter(cls, pythons):
    """Filters duplicate python interpreters and versions we don't support.

    Returns an iterator over PythonInterpreters.
    """
    MAJOR, MINOR, SUBMINOR = range(3)
    def version_filter(version):
      return (version[MAJOR] == 2 and version[MINOR] >= 7 or
              version[MAJOR] == 3 and version[MINOR] >= 5)

    seen = set()
    for interp in pythons:
      version = interp.identity.version
      if version not in seen and version_filter(version):
        seen.add(version)
        yield interp

  @classmethod
  def _sanitized_environment(cls, env=None):
    # N.B. This is merely a hack because sysconfig.py on the default OS X
    # installation of 2.7 breaks.
    env_copy = (env or os.environ).copy()
    env_copy.pop('MACOSX_DEPLOYMENT_TARGET', None)
    return env_copy

  def __init__(self, identity):
    """Construct a PythonInterpreter.

    You should probably use `PythonInterpreter.from_binary` instead.

    :param identity: The :class:`PythonIdentity` of the PythonInterpreter.
    """
    self._identity = identity
    self._binary = self._normalize_path(self.identity.binary)

    self._PYTHON_INTERPRETER_BY_NORMALIZED_PATH[self._binary] = self

  @property
  def binary(self):
    return self._binary

  @property
  def identity(self):
    return self._identity

  @property
  def python(self):
    return self._identity.python

  @property
  def version(self):
    return self._identity.version

  @property
  def version_string(self):
    return str(self._identity)

  def execute(self, args=None, stdin_payload=None, pythonpath=None, env=None, **kwargs):
    return self._execute(self.binary,
                         args=args,
                         stdin_payload=stdin_payload,
                         pythonpath=pythonpath,
                         env=env,
                         **kwargs)

  def open_process(self, args=None, pythonpath=None, env=None, **kwargs):
    cmd, env = self._create_isolated_cmd(self.binary, args=args, pythonpath=pythonpath, env=env)
    process = Executor.open_process(cmd, env=env, **kwargs)
    return cmd, process

  def _tup(self):
    return self._binary, self._identity

  def __hash__(self):
    return hash(self._tup())

  def __eq__(self, other):
    if type(other) is not type(self):
      return NotImplemented
    return self._tup() == other._tup()

  def __lt__(self, other):
    if type(other) is not type(self):
      return NotImplemented
    return self.version < other.version

  def __repr__(self):
    return '{type}({binary!r}, {identity!r})'.format(
      type=self.__class__.__name__,
      binary=self._binary,
      identity=self._identity
    )


def spawn_python_job(
  args,
  env=None,
  interpreter=None,
  expose=None,
  pythonpath=None,
  **subprocess_kwargs
):
  """Spawns a python job.

  :param args: The arguments to pass to the python interpreter.
  :type args: list of str
  :param env: The environment to spawn the python interpreter process in. Defaults to the ambient
              environment.
  :type env: dict of (str, str)
  :param interpreter: The interpreter to use to spawn the python job. Defaults to the current
                      interpreter.
  :type interpreter: :class:`PythonInterpreter`
  :param expose: The names of any vendored distributions to expose to the spawned python process.
                 These will be appended to `pythonpath` if passed.
  :type expose: list of str
  :param pythonpath: The PYTHONPATH to expose to the spawned python process. These will be
                     pre-pended to the `expose` path if passed.
  :type pythonpath: list of str
  :param subprocess_kwargs: Any additional :class:`subprocess.Popen` kwargs to pass through.
  :returns: A job handle to the spawned python process.
  :rtype: :class:`Job`
  """
  pythonpath = list(pythonpath or ())
  if expose:
    subprocess_env = (env or os.environ).copy()
    # In order to expose vendored distributions with their un-vendored import paths in-tact, we
    # need to set `__PEX_UNVENDORED__`. See: vendor.__main__.ImportRewriter._modify_import.
    subprocess_env['__PEX_UNVENDORED__'] = '1'

    pythonpath.extend(third_party.expose(expose))
  else:
    subprocess_env = env

  interpreter = interpreter or PythonInterpreter.get()
  cmd, process = interpreter.open_process(
    args=args,
    pythonpath=pythonpath,
    env=subprocess_env,
    **subprocess_kwargs
  )
  return Job(command=cmd, process=process)



# coding=utf-8
# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import warnings


class PEXWarning(Warning):
  """Indicates a warning from PEX about suspect buildtime or runtime configuration."""


def configure_warnings(pex_info, env):
  if env.PEX_VERBOSE > 0:
    emit_warnings = True
  elif env.PEX_EMIT_WARNINGS is not None:
    emit_warnings = env.PEX_EMIT_WARNINGS
  else:
    emit_warnings = pex_info.emit_warnings

  action = 'default' if emit_warnings else 'ignore'
  warnings.filterwarnings(action, category=PEXWarning)


def warn(message):
  warnings.warn(message, category=PEXWarning, stacklevel=2)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

import pytest

from pex.finders import get_entry_point_from_console_script, get_script_from_distributions
from pex.pip import get_pip
from pex.util import DistributionHelper


# In-part, tests a bug where the wheel distribution name has dashes as reported in:
#   https://github.com/pantsbuild/pex/issues/443
#   https://github.com/pantsbuild/pex/issues/551
def test_get_script_from_distributions(tmpdir):
  whl_path = './tests/example_packages/aws_cfn_bootstrap-1.4-py2-none-any.whl'
  install_dir = os.path.join(str(tmpdir), os.path.basename(whl_path))
  get_pip().spawn_install_wheel(wheel=whl_path, install_dir=install_dir).wait()

  dist = DistributionHelper.distribution_from_path(install_dir)
  assert 'aws-cfn-bootstrap' == dist.project_name

  dist_script = get_script_from_distributions('cfn-signal', [dist])
  assert dist_script.dist is dist
  assert os.path.join(install_dir, 'bin/cfn-signal') == dist_script.path
  assert dist_script.read_contents().startswith('#!'), (
    'Expected a `scripts`-style script w/shebang.'
  )

  assert None is get_script_from_distributions('non_existent_script', [dist])


class FakeDist(object):
  def __init__(self, key, console_script_entry):
    self.key = key
    script = console_script_entry.split('=')[0].strip()
    self._entry_map = {'console_scripts': {script: console_script_entry}}

  def get_entry_map(self):
    return self._entry_map


def test_get_entry_point_from_console_script():
  dists = [FakeDist(key='fake', console_script_entry='bob= bob.main:run'),
           FakeDist(key='fake', console_script_entry='bob =bob.main:run')]

  dist, entrypoint = get_entry_point_from_console_script('bob', dists)
  assert 'bob.main:run' == entrypoint
  assert dist in dists


def test_get_entry_point_from_console_script_conflict():
  dists = [FakeDist(key='bob', console_script_entry='bob= bob.main:run'),
           FakeDist(key='fake', console_script_entry='bob =bob.main:run')]
  with pytest.raises(RuntimeError):
    get_entry_point_from_console_script('bob', dists)


def test_get_entry_point_from_console_script_dne():
  dists = [FakeDist(key='bob', console_script_entry='bob= bob.main:run'),
           FakeDist(key='fake', console_script_entry='bob =bob.main:run')]
  assert (None, None) == get_entry_point_from_console_script('jane', dists)



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys
from contextlib import contextmanager

from pex import third_party
from pex.common import temporary_dir
from pex.variables import ENV


@contextmanager
def temporary_pex_root():
  with temporary_dir() as pex_root, ENV.patch(PEX_ROOT=os.path.realpath(pex_root)) as env:
    original_isolated = third_party._ISOLATED
    try:
      third_party._ISOLATED = None
      yield os.path.realpath(pex_root), env
    finally:
      third_party._ISOLATED = original_isolated


def test_isolated_pex_root():
  with temporary_pex_root() as (pex_root, _):
    devendored_chroot = os.path.realpath(third_party.isolated().chroot_path)
    assert pex_root == os.path.commonprefix([pex_root, devendored_chroot])


def test_isolated_idempotent_inprocess():
  with temporary_pex_root():
    result1 = third_party.isolated()
    result2 = third_party.isolated()
    assert result1.pex_hash == result2.pex_hash
    assert os.path.realpath(result1.chroot_path) == os.path.realpath(result2.chroot_path)


def test_isolated_idempotent_subprocess():
  with temporary_pex_root() as (_, env):
    devendored_chroot = os.path.realpath(third_party.isolated().chroot_path)
    stdout = subprocess.check_output(
      args=[
        sys.executable,
        '-c',
        'from pex import third_party; print(third_party.isolated().chroot_path)'
      ],
      env=env
    )
    assert devendored_chroot == os.path.realpath(stdout.decode('utf-8').strip())



# Copyright 2018 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import contextlib
import importlib
import os
import re
import shutil
import sys
import zipfile
from collections import OrderedDict, namedtuple


# NB: All pex imports are performed lazily to play well with the un-imports performed by both the
# PEX runtime when it demotes the bootstrap code and any pex modules that uninstalled
# VendorImporters un-import.
def _tracer():
  from pex.tracer import TRACER

  return TRACER


class _Loader(namedtuple('_Loader', ['module_name', 'vendor_module_name'])):

  # The PEP-302 loader API.
  # See: https://www.python.org/dev/peps/pep-0302/#specification-part-1-the-importer-protocol
  def load_module(self, fullname):
    assert fullname in (self.module_name, self.vendor_module_name), (
      '{} got an unexpected module {}'.format(self, fullname)
    )
    vendored_module = importlib.import_module(self.vendor_module_name)
    sys.modules[fullname] = vendored_module
    _tracer().log('{} imported via {}'.format(fullname, self), V=9)
    return vendored_module

  def unload(self):
    for mod in (self.module_name, self.vendor_module_name):
      if mod in sys.modules:
        sys.modules.pop(mod)
        _tracer().log('un-imported {}'.format(mod), V=9)

        submod_prefix = mod + '.'
        for submod in sorted(m for m in sys.modules.keys() if m.startswith(submod_prefix)):
          sys.modules.pop(submod)
          _tracer().log('un-imported {}'.format(submod), V=9)


class _Importable(namedtuple('_Importable', ['module', 'is_pkg', 'path', 'prefix'])):
  _exposed = False  # noqa: We want instance variable access defaulting to cls here.

  def expose(self):
    self._exposed = True
    importlib.import_module(self.module)
    _tracer().log('Exposed {}'.format(self), V=3)

  def loader_for(self, fullname):
    if fullname.startswith(self.prefix + '.'):
      target = fullname[len(self.prefix + '.'):]
    else:
      if not self._exposed:
        return None
      target = fullname

    if target == self.module or self.is_pkg and target.startswith(self.module + '.'):
      vendor_path = os.path.join(self.path, *target.split('.'))
      vendor_module_name = vendor_path.replace(os.sep, '.')
      return _Loader(fullname, vendor_module_name)


class _DirIterator(namedtuple('_DirIterator', ['rootdir'])):
  def iter_root_modules(self, relpath):
    for entry in self._iter_root(relpath):
      if os.path.isfile(entry):
        name, ext = os.path.splitext(os.path.basename(entry))
        if ext == '.py' and name != '__init__':
          yield name

  def iter_root_packages(self, relpath):
    for entry in self._iter_root(relpath):
      if os.path.isfile(os.path.join(entry, '__init__.py')):
        yield os.path.basename(entry)

  def _iter_root(self, relpath):
    root = os.path.join(self.rootdir, relpath)
    if not os.path.isdir(root):
      # We have nothing at this relpath as can happen when vendoring subsets of pex into its
      # runtime; ie: .bootstrap/pex gets pkg_resources but no setuptools or wheel.
      return

    for entry in os.listdir(root):
      yield os.path.join(root, entry)


class _ZipIterator(namedtuple('_ZipIterator', ['zipfile_path', 'prefix'])):
  @classmethod
  def containing(cls, root):
    prefix = ''
    path = root
    while path:
      # We use '/' here because the zip file format spec specifies that paths must use
      # forward slashes. See section 4.4.17 of
      # https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT.
      if zipfile.is_zipfile(path):
        return cls(zipfile_path=path, prefix='{}/'.format(prefix) if prefix else '')
      path_base = os.path.basename(path)
      prefix = '{}/{}'.format(path_base, prefix) if prefix else path_base
      path = os.path.dirname(path)
    raise ValueError('Could not find the zip file housing {}'.format(root))

  def iter_root_modules(self, relpath):
    for package in self._filter_names(relpath, r'(?P<module>[^/]+)\.py', 'module'):
      if package != '__init__':
        yield package

  def iter_root_packages(self, relpath):
    for package in self._filter_names(relpath, r'(?P<package>[^/]+)/__init__\.py', 'package'):
      yield package

  def _filter_names(self, relpath, pattern, group):
    # We use '/' here because the zip file format spec specifies that paths must use
    # forward slashes. See section 4.4.17 of
    # https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT.
    relpath_pat = '' if not relpath else '{}/'.format(relpath.replace(os.sep, '/'))
    pat = re.compile(r'^{}{}{}$'.format(self.prefix, relpath_pat, pattern))
    with contextlib.closing(zipfile.ZipFile(self.zipfile_path)) as zf:
      for name in zf.namelist():
        match = pat.match(name)
        if match:
          yield match.group(group)


class VendorImporter(object):
  """A `PEP-302 <https://www.python.org/dev/peps/pep-0302/>`_ meta_path importer for vendored code.

  This importer redirects imports from its package to vendored code, optionally exposing the
  vendored code by its un-prefixed module name as well.

  For example, if the ``requests`` distribution was vendored, it could be imported using this
  importer via ``import pex.third_party.requests`` as long as:

    * The requests distribution was housed under some importable path prefix inside this
      distribution.
    * The requests distribution had its self-referential absolute imports re-written to use the
      vendored import prefix.
  """

  @staticmethod
  def _abs_root(root=None):
    from pex import vendor

    return os.path.abspath(root or vendor.VendorSpec.ROOT)

  @classmethod
  def _iter_importables(cls, root, path_items, prefix):
    module_iterator = _DirIterator(root) if os.path.isdir(root) else _ZipIterator.containing(root)
    for path_item in path_items:
      for module_name in module_iterator.iter_root_modules(path_item):
        yield _Importable(module=module_name, is_pkg=False, path=path_item, prefix=prefix)
      for package_name in module_iterator.iter_root_packages(path_item):
        yield _Importable(module=package_name, is_pkg=True, path=path_item, prefix=prefix)

  @classmethod
  def _iter_all_installed_vendor_importers(cls):
    for importer in sys.meta_path:
      if isinstance(importer, cls):
        yield importer

  @classmethod
  def _iter_installed_vendor_importers(cls, prefix, root, path_items):
    for importer in cls._iter_all_installed_vendor_importers():
      # All Importables for a given VendorImporter will have the same prefix.
      if importer._importables and importer._importables[0].prefix == prefix:
        if importer._root == root:
          if {importable.path for importable in importer._importables} == set(path_items):
            yield importer

  @classmethod
  def install_vendored(cls, prefix, root=None, expose=None):
    """Install an importer for all vendored code with the given import prefix.

    All distributions listed in ``expose`` will also be made available for import in direct,
    un-prefixed form.

    :param str prefix: The import prefix the installed importer will be responsible for.
    :param str root: The root path of the distribution containing the vendored code. NB: This is the
                     the path to the pex code, which serves as the root under which code is vendored
                     at ``pex/vendor/_vendored``.
    :param expose: Optional names of distributions to expose for direct, un-prefixed import.
    :type expose: list of str
    :raise: :class:`ValueError` if any distributions to expose cannot be found.
    """
    from pex import vendor

    root = cls._abs_root(root)
    vendored_path_items = [spec.relpath for spec in vendor.iter_vendor_specs()]

    installed = list(cls._iter_installed_vendor_importers(prefix, root, vendored_path_items))
    assert len(installed) <= 1, (
      'Unexpected extra importers installed for vendored code:\n\t{}'
        .format('\n\t'.join(map(str, installed)))
    )
    if installed:
      vendor_importer = installed[0]
    else:
      # Install all vendored code for pex internal access to it through the vendor import `prefix`.
      vendor_importer = cls.install(uninstallable=True,
                                    prefix=prefix,
                                    path_items=vendored_path_items,
                                    root=root)

    if expose:
      # But only expose the bits needed.
      exposed_paths = []
      for path in cls.expose(expose, root):
        sys.path.insert(0, path)
        exposed_paths.append(os.path.relpath(path, root))

      vendor_importer._expose(exposed_paths)

  @classmethod
  def expose(cls, dists, root=None):
    from pex import vendor

    root = cls._abs_root(root)

    def iter_available():
      yield 'pex', root  # The pex distribution itself is trivially available to expose.
      for spec in vendor.iter_vendor_specs():
        yield spec.key, spec.relpath

    path_by_key = OrderedDict((key, relpath) for key, relpath in iter_available() if key in dists)

    unexposed = set(dists) - set(path_by_key.keys())
    if unexposed:
      raise ValueError('The following vendored dists are not available to expose: {}'
                       .format(', '.join(sorted(unexposed))))

    exposed_paths = path_by_key.values()
    for exposed_path in exposed_paths:
      yield os.path.join(root, exposed_path)

  @classmethod
  def install(cls, uninstallable, prefix, path_items, root=None, warning=None):
    """Install an importer for modules found under ``path_items`` at the given import ``prefix``.

    :param bool uninstallable: ``True`` if the installed importer should be uninstalled and any
                               imports it performed be un-imported when ``uninstall`` is called.
    :param str prefix: The import prefix the installed importer will be responsible for.
    :param path_items: The paths relative to ``root`` containing modules to expose for import under
                       ``prefix``.
    :param str root: The root path of the distribution containing the vendored code. NB: This is the
                     the path to the pex code, which serves as the root under which code is vendored
                     at ``pex/vendor/_vendored``.
    :param str warning: An optional warning to emit if any imports are made through the installed
                        importer.
    :return:
    """
    root = cls._abs_root(root)
    importables = tuple(cls._iter_importables(root=root, path_items=path_items, prefix=prefix))
    vendor_importer = cls(root=root,
                          importables=importables,
                          uninstallable=uninstallable,
                          warning=warning)
    sys.meta_path.insert(0, vendor_importer)
    _tracer().log('Installed {}'.format(vendor_importer), V=3)
    return vendor_importer

  @classmethod
  def uninstall_all(cls):
    """Uninstall all uninstallable VendorImporters and unimport the modules they loaded."""
    for vendor_importer in cls._iter_all_installed_vendor_importers():
      vendor_importer.uninstall()

  def __init__(self, root, importables, uninstallable=True, warning=None):
    self._root = root
    self._importables = importables

    self._uninstallable = uninstallable
    self._warning = warning

    self._loaders = []

  def uninstall(self):
    """Uninstall this importer if possible and un-import any modules imported by it."""
    if not self._uninstallable:
      _tracer().log('Not uninstalling {}'.format(self), V=9)
      return

    if self in sys.meta_path:
      sys.meta_path.remove(self)
      maybe_exposed = frozenset(os.path.join(self._root, importable.path)
                                for importable in self._importables)
      sys.path[:] = [path_item for path_item in sys.path if path_item not in maybe_exposed]
      for loader in self._loaders:
        loader.unload()
      _tracer().log('Uninstalled {}'.format(self), V=3)

  # The PEP-302 finder API.
  # See: https://www.python.org/dev/peps/pep-0302/#specification-part-1-the-importer-protocol
  def find_module(self, fullname, path=None):
    for importable in self._importables:
      loader = importable.loader_for(fullname)
      if loader is not None:
        self._loaders.append(loader)
        if self._warning:
          from pex import pex_warnings
          pex_warnings.warn('Found loader for `import {}`:\n\t{}'.format(fullname, self._warning))
        return loader
    return None

  def _expose(self, paths):
    for importable in self._importables:
      if importable.path in paths:
        importable.expose()

  def __repr__(self):
    return ('{classname}(root={root!r}, importables={importables!r})'
            .format(classname=self.__class__.__name__,
                    root=self._root,
                    importables=self._importables))


class IsolationResult(namedtuple('IsolatedPex', ['pex_hash', 'chroot_path'])):
  """The result of isolating the current pex distribution to a filesystem chroot."""


_ISOLATED = None


def isolated():
  """Returns a chroot for third_party isolated from the ``sys.path``.

  PEX will typically be installed in site-packages flat alongside many other distributions; as such,
  adding the location of the pex distribution to the ``sys.path`` will typically expose many other
  distributions. An isolated chroot can be used as a ``sys.path`` entry to effect only the exposure
  of pex.

  :return: An isolation result.
  :rtype: :class:`IsolationResult`
  """
  global _ISOLATED
  if _ISOLATED is None:
    from pex import vendor
    from pex.common import atomic_directory
    from pex.util import CacheHelper
    from pex.variables import ENV
    from pex.third_party.pkg_resources import resource_isdir, resource_listdir, resource_stream

    module = 'pex'
    def recursive_copy(srcdir, dstdir):
      os.mkdir(dstdir)
      for entry_name in resource_listdir(module, srcdir):
        # NB: Resource path components are always separated by /, on all systems.
        src_entry = '{}/{}'.format(srcdir, entry_name) if srcdir else entry_name
        dst_entry = os.path.join(dstdir, entry_name)
        if resource_isdir(module, src_entry):
          recursive_copy(src_entry, dst_entry)
        elif not entry_name.endswith('.pyc'):
          with open(dst_entry, 'wb') as fp:
            shutil.copyfileobj(resource_stream(module, src_entry), fp)

    pex_path = os.path.join(vendor.VendorSpec.ROOT, 'pex')
    with _tracer().timed('Hashing pex'):
      dir_hash = CacheHelper.dir_hash(pex_path)
    isolated_dir = os.path.join(ENV.PEX_ROOT, 'isolated', dir_hash)

    with _tracer().timed('Isolating pex'):
      with atomic_directory(isolated_dir) as chroot:
        if chroot:
          with _tracer().timed('Extracting pex to {}'.format(isolated_dir)):
            recursive_copy('', os.path.join(chroot, 'pex'))

    _ISOLATED = IsolationResult(pex_hash=dir_hash, chroot_path=isolated_dir)
  return _ISOLATED


def uninstall():
  """Uninstall all uninstallable :class:`VendorImporter`s and uninmport the modules they loaded."""
  VendorImporter.uninstall_all()


def import_prefix():
  """Returns the vendoring import prefix; eg: `pex.third_party`.

  :rtype: str
  """
  return __name__


def install(root=None, expose=None):
  """Installs the default :class:`VendorImporter` for PEX vendored code.

  Any distributions listed in ``expose`` will also be exposed for direct import; ie:
  ``install(expose=['setuptools'])`` would make both ``setuptools`` and ``wheel`` available for
  import via ``from  pex.third_party import setuptools, wheel``, but only ``setuptools`` could be
  directly imported via ``import setuptools``.

  NB: Even when exposed, vendored code is not the same as the same un-vendored code and will
  properly fail type-tests against un-vendored types. For example, in an interpreter that has
  ``setuptools`` installed in its site-packages:

  >>> from pkg_resources import Requirement
  >>> orig_req = Requirement.parse('wheel==0.31.1')
  >>> from pex import third_party
  >>> third_party.install(expose=['setuptools'])
  >>> import sys
  >>> sys.modules.pop('pkg_resources')
  <module 'pkg_resources' from '/home/jsirois/dev/pantsbuild/jsirois-pex/.tox/py27-repl/lib/python2.7/site-packages/pkg_resources/__init__.pyc'>  # noqa
  >>> from pkg_resources import Requirement
  >>> new_req = Requirement.parse('wheel==0.31.1')
  >>> new_req == orig_req
  False
  >>> new_req == Requirement.parse('wheel==0.31.1')
  True
  >>> type(orig_req)
  <class 'pkg_resources.Requirement'>
  >>> type(new_req)
  <class 'pex.vendor._vendored.setuptools.pkg_resources.Requirement'>
  >>> from pex.third_party.pkg_resources import Requirement as PrefixedRequirement
  >>> new_req == PrefixedRequirement.parse('wheel==0.31.1')
  True
  >>> sys.modules.pop('pkg_resources')
  <module 'pex.vendor._vendored.setuptools.pkg_resources' from 'pex/vendor/_vendored/setuptools/pkg_resources/__init__.pyc'>  # noqa
  >>> sys.modules.pop('pex.third_party.pkg_resources')
  <module 'pex.vendor._vendored.setuptools.pkg_resources' from 'pex/vendor/_vendored/setuptools/pkg_resources/__init__.pyc'>  # noqa
  >>>

  :param expose: A list of vendored distribution names to expose directly on the ``sys.path``.
  :type expose: list of str
  :raise: :class:`ValueError` if any distributions to expose cannot be found.
  """
  VendorImporter.install_vendored(prefix=import_prefix(), root=root, expose=expose)


def expose(dists):
  """Exposes vendored code in isolated chroots.

  Any vendored distributions listed in ``dists`` will be unpacked to individual chroots for addition
  to the ``sys.path``; ie: ``expose(['setuptools', 'wheel'])`` will unpack these vendored
  distributions and yield the two chroot paths they were unpacked to.

  :param dists: A list of vendored distribution names to expose.
  :type dists: list of str
  :raise: :class:`ValueError` if any distributions to expose cannot be found.
  :returns: An iterator of exposed vendored distribution chroot paths.
  """
  for path in VendorImporter.expose(dists, root=isolated().chroot_path):
    yield path


# Implicitly install an importer for vendored code on the first import of pex.third_party.
install()



# Copyright 2018 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import collections
import os

from pex.common import touch
from pex.compatibility import urlparse
from pex.tracer import TRACER

_PACKAGE_COMPONENTS = __name__.split('.')


def _root():
  path = os.path.dirname(os.path.abspath(__file__))
  for _ in _PACKAGE_COMPONENTS:
    path = os.path.dirname(path)
  return path


class VendorSpec(collections.namedtuple('VendorSpec',
                                        ['key', 'requirement', 'rewrite', 'constrain'])):
  """Represents a vendored distribution.

  :field str key: The distribution requirement key; e.g.: for a requirement of
    requests[security]==2.22.0 the key is 'requests'.
  :field str requirement: The distribution requirement string; e.g.: requests[security]==2.22.0.
  :field bool rewrite: Whether to re-write the distribution's imports for use with the
    `pex.third_party` importer.
  :field bool constrain: Whether to attempt to constrain the requirement via pip's --constraint
    mechanism.

  NB: Vendored distributions should comply with the host distribution platform constraints. In the
  case of pex, which is a py2.py3 platform agnostic wheel, vendored libraries should be as well.
  """

  ROOT = _root()

  @classmethod
  def pinned(cls, key, version, rewrite=True):
    return cls(key=key, requirement='{}=={}'.format(key, version), rewrite=rewrite, constrain=True)

  @classmethod
  def vcs(cls, url, rewrite=True):
    result = urlparse.urlparse(url)
    fragment_params = urlparse.parse_qs(result.fragment)
    values = fragment_params.get('egg')
    if not values or len(values) != 1:
      raise ValueError('Expected the vcs requirement url to have an #egg=<name> fragment. '
                       'Got: {}'.format(url))

    # N.B.: Constraints do not work for vcs urls.
    return cls(key=values[0], requirement=url, rewrite=rewrite, constrain=False)

  @property
  def _subpath_components(self):
    return ['_vendored', self.key]

  @property
  def relpath(self):
    return os.path.join(*(_PACKAGE_COMPONENTS + self._subpath_components))

  @property
  def target_dir(self):
    return os.path.join(self.ROOT, self.relpath)

  def create_packages(self):
    """Create missing packages joining the vendor root to the base of the vendored distribution.

    For example, given a root at ``/home/jake/dev/pantsbuild/pex`` and a vendored distribution at
    ``pex/vendor/_vendored/requests`` this method would create the following package files::

      pex/vendor/_vendored/__init__.py
      pex/vendor/_vendored/requests/__init__.py

    These package files allow for standard python importers to find vendored code via re-directs
    from a `PEP-302 <https://www.python.org/dev/peps/pep-0302/>`_ importer like
    :class:`pex.third_party.VendorImporter`.
    """
    if not self.rewrite:
      # The extra package structure is only required for vendored code used via import rewrites.
      return

    for index, _ in enumerate(self._subpath_components):
      relpath = _PACKAGE_COMPONENTS + self._subpath_components[:index + 1] + ['__init__.py']
      touch(os.path.join(self.ROOT, *relpath))


def iter_vendor_specs():
  """Iterate specifications for code vendored by pex.

  :return: An iterator over specs of all vendored code.
  :rtype: :class:`collection.Iterator` of :class:`VendorSpec`
  """
  # We use this via pex.third_party at runtime to check for compatible wheel tags.
  yield VendorSpec.pinned('packaging', '19.2')

  # We shell out to pip at buildtime to resolve and install dependencies.
  # N.B.: This is pip 20.0.dev0 with a patch to support foreign download targets more fully.
  yield VendorSpec.vcs('git+https://github.com/pantsbuild/pip@5eb9470c0c59#egg=pip', rewrite=False)

  # We expose this to pip at buildtime for legacy builds, but we also use pkg_resources via
  # pex.third_party at runtime in various ways.
  yield VendorSpec.pinned('setuptools', '42.0.2')

  # We expose this to pip at buildtime for legacy builds.
  yield VendorSpec.pinned('wheel', '0.33.6', rewrite=False)


def vendor_runtime(chroot, dest_basedir, label, root_module_names):
  """Includes portions of vendored distributions in a chroot.

  The portion to include is selected by root module name. If the module is a file, just it is
  included. If the module represents a package, the package and all its sub-packages are added
  recursively.

  :param chroot: The chroot to add vendored code to.
  :type chroot: :class:`pex.common.Chroot`
  :param str dest_basedir: The prefix to store the vendored code under in the ``chroot``.
  :param str label: The chroot label for the vendored code fileset.
  :param root_module_names: The names of the root vendored modules to include in the chroot.
  :type root_module_names: :class:`collections.Iterable` of str
  :raise: :class:`ValueError` if any of the given ``root_module_names`` could not be found amongst
          the vendored code and added to the chroot.
  """
  vendor_module_names = {root_module_name: False for root_module_name in root_module_names}
  for spec in iter_vendor_specs():
    for root, dirs, files in os.walk(spec.target_dir):
      if root == spec.target_dir:
        dirs[:] = [pkg_name for pkg_name in dirs if pkg_name in vendor_module_names]
        files[:] = [mod_name for mod_name in files if mod_name[:-3] in vendor_module_names]
        vendored_names = dirs + [filename[:-3] for filename in files]
        if vendored_names:
          pkg_path = ''
          for pkg in spec.relpath.split(os.sep):
            pkg_path = os.path.join(pkg_path, pkg)
            pkg_file = os.path.join(pkg_path, '__init__.py')
            src = os.path.join(VendorSpec.ROOT, pkg_file)
            dest = os.path.join(dest_basedir, pkg_file)
            if os.path.exists(src):
              chroot.copy(src, dest, label)
            else:
              # We delete `pex/vendor/_vendored/<dist>/__init__.py` when isolating third_party.
              chroot.touch(dest, label)
          for name in vendored_names:
            vendor_module_names[name] = True
            TRACER.log('Vendoring {} from {} @ {}'.format(name, spec, spec.target_dir), V=3)

      for filename in files:
        if not filename.endswith('.pyc'):  # Sources and data only.
          src = os.path.join(root, filename)
          dest = os.path.join(dest_basedir, spec.relpath, os.path.relpath(src, spec.target_dir))
          chroot.copy(src, dest, label)

  if not all(vendor_module_names.values()):
    raise ValueError('Failed to extract {module_names} from:\n\t{specs}'.format(
      module_names=', '.join(module
                             for module, written in vendor_module_names.items() if not written),
      specs='\n\t'.join('{} @ {}'.format(spec, spec.target_dir) for spec in iter_vendor_specs())))



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import functools
import os
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_copy, safe_mkdtemp, temporary_dir
from pex.compatibility import nested
from pex.interpreter import PythonInterpreter
from pex.resolver import resolve_multi
from pex.testing import (
    IS_LINUX,
    IS_PYPY,
    PY27,
    PY35,
    PY36,
    PY_VER,
    built_wheel,
    ensure_python_interpreter,
    make_source_dir
)
from pex.third_party.pkg_resources import Requirement


def build_wheel(*args, **kwargs):
  with built_wheel(*args, **kwargs) as whl:
    return whl


def local_resolve_multi(*args, **kwargs):
  # Skip remote lookups.
  kwargs['indexes'] = []
  return list(resolve_multi(*args, **kwargs))


def test_empty_resolve():
  empty_resolve_multi = local_resolve_multi([])
  assert empty_resolve_multi == []

  with temporary_dir() as td:
    empty_resolve_multi = local_resolve_multi([], cache=td)
    assert empty_resolve_multi == []


def test_simple_local_resolve():
  project_wheel = build_wheel(name='project')

  with temporary_dir() as td:
    safe_copy(project_wheel, os.path.join(td, os.path.basename(project_wheel)))
    resolved_dists = local_resolve_multi(['project'], find_links=[td])
    assert len(resolved_dists) == 1


def test_resolve_cache():
  project_wheel = build_wheel(name='project')

  with nested(temporary_dir(), temporary_dir()) as (td, cache):
    safe_copy(project_wheel, os.path.join(td, os.path.basename(project_wheel)))

    # Without a cache, each resolve should be isolated, but otherwise identical.
    resolved_dists1 = local_resolve_multi(['project'], find_links=[td])
    resolved_dists2 = local_resolve_multi(['project'], find_links=[td])
    assert resolved_dists1 != resolved_dists2
    assert len(resolved_dists1) == 1
    assert len(resolved_dists2) == 1
    assert resolved_dists1[0].requirement == resolved_dists2[0].requirement
    assert resolved_dists1[0].distribution.location != resolved_dists2[0].distribution.location

    # With a cache, each resolve should be identical.
    resolved_dists3 = local_resolve_multi(['project'], find_links=[td], cache=cache)
    resolved_dists4 = local_resolve_multi(['project'], find_links=[td], cache=cache)
    assert resolved_dists1 != resolved_dists3
    assert resolved_dists2 != resolved_dists3
    assert resolved_dists3 == resolved_dists4


def test_diamond_local_resolve_cached():
  # This exercises the issue described here: https://github.com/pantsbuild/pex/issues/120
  project1_wheel = build_wheel(name='project1', install_reqs=['project2<1.0.0'])
  project2_wheel = build_wheel(name='project2')

  with temporary_dir() as dd:
    for wheel in (project1_wheel, project2_wheel):
      safe_copy(wheel, os.path.join(dd, os.path.basename(wheel)))
    with temporary_dir() as cd:
      resolved_dists = local_resolve_multi(['project1', 'project2'],
                                           find_links=[dd],
                                           cache=cd)
      assert len(resolved_dists) == 2


def test_cached_dependency_pinned_unpinned_resolution_multi_run():
  # This exercises the issue described here: https://github.com/pantsbuild/pex/issues/178
  project1_0_0 = build_wheel(name='project', version='1.0.0')
  project1_1_0 = build_wheel(name='project', version='1.1.0')

  with temporary_dir() as td:
    for wheel in (project1_0_0, project1_1_0):
      safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))
    with temporary_dir() as cd:
      # First run, pinning 1.0.0 in the cache
      resolved_dists = local_resolve_multi(['project==1.0.0'],
                                           find_links=[td],
                                           cache=cd)
      assert len(resolved_dists) == 1
      assert resolved_dists[0].distribution.version == '1.0.0'

      # Second, run, the unbounded 'project' req will find the 1.0.0 in the cache. But should also
      # return SourcePackages found in td
      resolved_dists = local_resolve_multi(['project'],
                                           find_links=[td],
                                           cache=cd)
      assert len(resolved_dists) == 1
      assert resolved_dists[0].distribution.version == '1.1.0'


def test_intransitive():
  foo1_0 = build_wheel(name='foo', version='1.0.0')
  # The nonexistent req ensures that we are actually not acting transitively (as that would fail).
  bar1_0 = build_wheel(name='bar', version='1.0.0', install_reqs=['nonexistent==1.0.0'])
  with temporary_dir() as td:
    for wheel in (foo1_0, bar1_0):
      safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))
    with temporary_dir() as cd:
      resolved_dists = local_resolve_multi(['foo', 'bar'],
                                           find_links=[td],
                                           cache=cd,
                                           transitive=False)
      assert len(resolved_dists) == 2


def test_resolve_prereleases():
  stable_dep = build_wheel(name='dep', version='2.0.0')
  prerelease_dep = build_wheel(name='dep', version='3.0.0rc3')

  with temporary_dir() as td:
    for wheel in (stable_dep, prerelease_dep):
      safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))

    def assert_resolve(expected_version, **resolve_kwargs):
      resolved_dists = local_resolve_multi(['dep>=1,<4'], find_links=[td], **resolve_kwargs)
      assert 1 == len(resolved_dists)
      resolved_dist = resolved_dists[0]
      assert expected_version == resolved_dist.distribution.version

    assert_resolve('2.0.0')
    assert_resolve('2.0.0', allow_prereleases=False)
    assert_resolve('3.0.0rc3', allow_prereleases=True)


def _parse_requirement(req):
  return Requirement.parse(str(req))


def test_resolve_extra_setup_py():
  with make_source_dir(name='project1',
                       version='1.0.0',
                       extras_require={'foo': ['project2']}) as project1_dir:
    project2_wheel = build_wheel(name='project2', version='2.0.0')
    with temporary_dir() as td:
      safe_copy(project2_wheel, os.path.join(td, os.path.basename(project2_wheel)))

      resolved_dists = local_resolve_multi(['{}[foo]'.format(project1_dir)], find_links=[td])
      assert ({_parse_requirement(req) for req in ('project1==1.0.0',
                                                   'project2==2.0.0')} ==
              {_parse_requirement(resolved_dist.requirement) for resolved_dist in resolved_dists})


def test_resolve_extra_wheel():
  project1_wheel = build_wheel(name='project1',
                               version='1.0.0',
                               extras_require={'foo': ['project2']})
  project2_wheel = build_wheel(name='project2', version='2.0.0')
  with temporary_dir() as td:
    for wheel in (project1_wheel, project2_wheel):
      safe_copy(wheel, os.path.join(td, os.path.basename(wheel)))

    resolved_dists = local_resolve_multi(['project1[foo]'], find_links=[td])
    assert ({_parse_requirement(req) for req in ('project1==1.0.0', 'project2==2.0.0')} ==
            {_parse_requirement(resolved_dist.requirement) for resolved_dist in resolved_dists})


def resolve_wheel_names(**kwargs):
  return [
    os.path.basename(resolved_distribution.distribution.location)
    for resolved_distribution in resolve_multi(**kwargs)
  ]


def resolve_p537_wheel_names(**kwargs):
  return resolve_wheel_names(requirements=['p537==1.0.4'], transitive=False, **kwargs)


@pytest.fixture(scope="module")
def p537_resolve_cache():
  return safe_mkdtemp()


@pytest.mark.skipif(PY_VER < (3, 5) or IS_PYPY,
                    reason="The p537 distribution only builds for CPython 3.5+")
def test_resolve_current_platform(p537_resolve_cache):
  resolve_current = functools.partial(resolve_p537_wheel_names,
                                      cache=p537_resolve_cache,
                                      platforms=['current'])

  other_python_version = PY36 if PY_VER == (3, 5) else PY35
  other_python = PythonInterpreter.from_binary(ensure_python_interpreter(other_python_version))
  current_python = PythonInterpreter.get()

  resolved_other = resolve_current(interpreters=[other_python])
  resolved_current = resolve_current()

  assert 1 == len(resolved_other)
  assert 1 == len(resolved_current)
  assert resolved_other != resolved_current
  assert resolved_current == resolve_current(interpreters=[current_python])
  assert resolved_current == resolve_current(interpreters=[current_python, current_python])

  # Here we have 2 local interpreters satisfying current but with different platforms and thus
  # different dists for 2 total dists.
  assert 2 == len(resolve_current(interpreters=[current_python, other_python]))


@pytest.mark.skipif(PY_VER < (3, 5) or IS_PYPY,
                    reason="The p537 distribution only builds for CPython 3.5+")
def test_resolve_current_and_foreign_platforms(p537_resolve_cache):
  foreign_platform = 'macosx-10.13-x86_64-cp-37-m' if IS_LINUX else 'manylinux1_x86_64-cp-37-m'
  resolve_current_and_foreign = functools.partial(resolve_p537_wheel_names,
                                                  cache=p537_resolve_cache,
                                                  platforms=['current', foreign_platform])

  assert 2 == len(resolve_current_and_foreign())

  other_python_version = PY36 if PY_VER == (3, 5) else PY35
  other_python = PythonInterpreter.from_binary(ensure_python_interpreter(other_python_version))
  current_python = PythonInterpreter.get()

  assert 2 == len(resolve_current_and_foreign(interpreters=[current_python]))
  assert 2 == len(resolve_current_and_foreign(interpreters=[other_python]))
  assert 2 == len(resolve_current_and_foreign(interpreters=[current_python, current_python]))

  # Here we have 2 local interpreters, satisfying current, but with different platforms and thus
  # different dists and then the foreign platform for 3 total dists.
  assert 3 == len(resolve_current_and_foreign(interpreters=[current_python, other_python]))


def test_resolve_foreign_abi3():
  # For version 2.8, cryptography publishes the following abi3 wheels for linux and macosx:
  # cryptography-2.8-cp34-abi3-macosx_10_6_intel.whl
  # cryptography-2.8-cp34-abi3-manylinux1_x86_64.whl
  # cryptography-2.8-cp34-abi3-manylinux2010_x86_64.whl

  cryptogrpahy_resolve_cache = safe_mkdtemp()
  foreign_ver = '37' if PY_VER == (3, 6) else '36'
  resolve_cryptography_wheel_names = functools.partial(
    resolve_wheel_names,
    requirements=['cryptography==2.8'],
    platforms=[
      'linux_x86_64-cp-{}-m'.format(foreign_ver),
      'macosx_10.11_x86_64-cp-{}-m'.format(foreign_ver)
    ],
    transitive=False,
    build=False,
    cache=cryptogrpahy_resolve_cache
  )

  wheel_names = resolve_cryptography_wheel_names(manylinux='manylinux2014')
  assert {
    'cryptography-2.8-cp34-abi3-manylinux2010_x86_64.whl',
    'cryptography-2.8-cp34-abi3-macosx_10_6_intel.whl'
  } == set(wheel_names)

  wheel_names = resolve_cryptography_wheel_names(manylinux='manylinux2010')
  assert {
    'cryptography-2.8-cp34-abi3-manylinux2010_x86_64.whl',
    'cryptography-2.8-cp34-abi3-macosx_10_6_intel.whl'
  } == set(wheel_names)

  wheel_names = resolve_cryptography_wheel_names(manylinux='manylinux1')
  assert {
    'cryptography-2.8-cp34-abi3-manylinux1_x86_64.whl',
    'cryptography-2.8-cp34-abi3-macosx_10_6_intel.whl'
  } == set(wheel_names)


def test_issues_851():
  # Previously, the PY36 resolve would fail post-resolution checks for configparser, pathlib2 and
  # contextlib2 which are only required for python_version<3.

  def resolve_pytest(python_version, pytest_version):
    interpreter = PythonInterpreter.from_binary(ensure_python_interpreter(python_version))
    resolved_dists = resolve_multi(interpreters=[interpreter],
                                           requirements=['pytest=={}'.format(pytest_version)])
    project_to_version = {rd.requirement.key: rd.distribution.version for rd in resolved_dists}
    assert project_to_version['pytest'] == pytest_version
    return project_to_version

  resolved_project_to_version = resolve_pytest(python_version=PY36, pytest_version='5.3.4')
  assert 'importlib-metadata' in resolved_project_to_version
  assert 'configparser' not in resolved_project_to_version
  assert 'pathlib2' not in resolved_project_to_version
  assert 'contextlib2' not in resolved_project_to_version

  resolved_project_to_version = resolve_pytest(python_version=PY27, pytest_version='4.6.9')
  assert 'importlib-metadata' in resolved_project_to_version
  assert 'configparser' in resolved_project_to_version
  assert 'pathlib2' in resolved_project_to_version
  assert 'contextlib2' in resolved_project_to_version


def test_issues_892():
  python27 = ensure_python_interpreter(PY27)
  program = dedent("""\
    from __future__ import print_function

    import os
    import sys


    # This puts python3.6 stdlib on PYTHONPATH.
    os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)


    from pex import resolver
    from pex.interpreter import PythonInterpreter


    python27 = PythonInterpreter.from_binary({python27!r})
    result = resolver.resolve(requirements=['packaging==19.2'], interpreter=python27)
    print('Resolved: {{}}'.format(result))
  """.format(python27=python27))

  python36 = ensure_python_interpreter(PY36)
  cmd, process = PythonInterpreter.from_binary(python36).open_process(
    args=['-c', program],
    stderr=subprocess.PIPE
  )
  _, stderr = process.communicate()
  assert process.returncode == 0, dedent(
    """
    Command {cmd} failed with {returncode}.

    STDERR
    ======
    {stderr}
    """.format(cmd=cmd, returncode=process.returncode, stderr=stderr.decode('utf8'))
  )
